#include "transform.h"
#include "object.h"
#include "ttransform.h"
#include <math.h>

Matrix::Matrix():
 r1(1),r2(0),r3(0),tx(0),
 r4(0),r5(1),r6(0),ty(0),
 r7(0),r8(0),r9(1),tz(0)
{}

Matrix operator* (const Matrix& a, const Matrix& b)
{
/* diagram to help with computing this matrix multiply
[ c1 c2 c3 cx ]   [ a1 a2 a3 ax ] [ b1 b2 b3 bx ]
[ c4 c5 c6 cy ]   [ a4 a5 a6 ay ] [ b4 b5 b6 by ]
[ c7 c8 c9 cz ] = [ a7 a8 a9 az ].[ b7 b8 b9 bz ]
[ 0  0  0  1  ]   [ 0  0  0  1  ] [ 0  0  0  1  ]  */
	Matrix c;
	c.r1=a.r1*b.r1 + a.r2*b.r4 + a.r3*b.r7;
	c.r2=a.r1*b.r2 + a.r2*b.r5 + a.r3*b.r8;
	c.r3=a.r1*b.r3 + a.r2*b.r6 + a.r3*b.r9;

	c.r4=a.r4*b.r1 + a.r5*b.r4 + a.r6*b.r7;
	c.r5=a.r4*b.r2 + a.r5*b.r5 + a.r6*b.r8;
	c.r6=a.r4*b.r3 + a.r5*b.r6 + a.r6*b.r9;

	c.r7=a.r7*b.r1 + a.r8*b.r4 + a.r9*b.r7;
	c.r8=a.r7*b.r2 + a.r8*b.r5 + a.r9*b.r8;
	c.r9=a.r7*b.r3 + a.r8*b.r6 + a.r9*b.r9;

	c.tx=a.r1*b.tx + a.r2*b.ty + a.r3*b.tz + a.tx;
	c.ty=a.r4*b.tx + a.r5*b.ty + a.r6*b.tz + a.ty;
	c.tz=a.r7*b.tx + a.r8*b.ty + a.r9*b.tz + a.tz;

	return c;
}

// Transformations

Vector transform_normalT(const Matrix& M, const Vector& b)
{ // when transforming the normal, use transpose of M
	Vector a;
	a.x=M.r1*b.x + M.r4*b.y + M.r7*b.z;
	a.y=M.r2*b.x + M.r5*b.y + M.r8*b.z;
	a.z=M.r3*b.x + M.r6*b.y + M.r9*b.z;
	return a;
}


void Transformation::rotateX(Float theta)
{
	Matrix M;
	Float c=cos(theta),s=sin(theta);
	M.r5=c; M.r6=-s; M.r8=s; M.r9=c;
	trans=M*trans;
	c=cos(-theta),s=sin(-theta);
	M.r5=c; M.r6=-s; M.r8=s; M.r9=c;
	inverse=inverse*M;
}

void Transformation::rotateY(Float theta)
{
	Matrix M;
	Float c=cos(theta),s=sin(theta);
	M.r1=c; M.r3=s; M.r7=-s; M.r9=c;
	trans=M*trans;
	c=cos(-theta),s=sin(-theta);
	M.r1=c; M.r3=s; M.r7=-s; M.r9=c;
	inverse=inverse*M;
}
void Transformation::rotateZ(Float theta)
{
	Matrix M;
	Float c=cos(theta),s=sin(theta);
	M.r1=c; M.r2=-s; M.r4=s; M.r5=c;
	trans=M*trans;
	c=cos(-theta);s=sin(-theta);
	M.r1=c; M.r2=-s; M.r4=s; M.r5=c;
	inverse=inverse*M;
}

void Transformation::translate(Vector T)
{
	Matrix M;
	M.tx=T.x; M.ty=T.y; M.tz=T.z;
	trans=M*trans;
	M.tx=-T.x; M.ty=-T.y; M.tz=-T.z;
	inverse=inverse*M;
}

void Transformation::scale(Vector S)
{
	Matrix M;
	M.r1=S.x; M.r5=S.y; M.r9=S.z;
	trans=M*trans;
	M.r1=1.0/S.x; M.r5=1.0/S.y; M.r9=1.0/S.z;
	inverse=inverse*M;
}

void Transformation::post_mult(Transformation *T)
//  A.B is easy enough.
//  (A.B)~=B~.A~
{
	Matrix M;
	M=trans*T->trans;
	trans=M;
	M=T->inverse * inverse;
	inverse=M;
}

Transformed_Object::Transformed_Object(Object* o,Transformation* t,Surface_Info &s):
	Object(s,o->t0,o->t1,transform),obj(o),T(*t)
{}

/* some notes about transforming from/to world/local co-ords.
    o  to transform a point from world co-ords to local,
       use the inverse matrix
    o use the trans matrix to transform a point from local to world.
*/
bool Transformed_Object::intersect(Ray& r,Float& t)
{
	Ray ray=r;
	ray.o=transform_point(T.inverse,r.o);
	ray.d=transform_normal(T.inverse,r.d);
	return obj->intersect(ray,t);
}

void Transformed_Object::aux_info(Ray& r,Intersect_Info& ii,Shade_Info& si)
{
	Ray ray;
	ray.o=transform_point(T.inverse,r.o);
	ray.d=transform_normal(T.inverse,r.d);
	obj->aux_info(ray,ii,si);
	// transform normal and point to world co-ords
	//si.P=transform_point(T.trans,si.P);
	si.P=r.o+r.d*ii.t;
	si.normal=transform_normalT(T.inverse,si.normal);
	si.normal.normalize();
}


Object* Transform_Object(Object* obj,Transformation* T,Surface_Info& si)
{
	if(obj->type==Object::primitive)
		return new Transformed_Object(obj,T,si);
	if(obj->type==Object::transform){
		Transformed_Object* Obj=(Transformed_Object*)(obj);
		Transformation* S= &Obj->T;
		Transformation* T1=new Transformation(*T);
		T1->post_mult(S);
		return new Transformed_Object(Obj->obj,T1,si);
	}
	if(obj->type==Object::ttransform){
		TTransformed_Object* Obj=(TTransformed_Object*)(obj);
		TTransformation* S=Obj->T;
		TTransformation* T1=new TTransformation(T,S->t0,S->t1);
		T1->post_mult(S);
		return new TTransformed_Object(Obj->obj,T1,si);
	}
}


